//-----------------------------------------------
// Copyright 2009 Wellcome Trust Sanger Institute
// Written by Jared Simpson (js18@sanger.ac.uk)
// Released under the GPL license
//-----------------------------------------------
import cnD;
import CNVModel;
import Common;
import Alphabet;
import Distribution;
import JointCNVDistribution;
import ModelParameters;
import StateSummary;
import tango.math.Math;
import tango.io.Stdout;

alias real[][char[]] ColumnArrayLUT;
real ZERO_DEPTH_PRIOR = 0.1;
real UNIQUE_RMEAN_PRIOR = 0.8;
real UNIQUE_RSD_PRIOR = 0.2;
real REPEAT_RMEAN_PRIOR = 0.1;
real REPEAT_RSD_PRIOR = 0.1;
real ZERO_RMEAN_PRIOR = 0.9;
real ZERO_RSD_PRIOR = 0.1;



ModelParameters generateInitialParameters(MarkovModel model, Symbol[] symbols, int windowSize)
{
	//
	// Initially, the states are fit using only the depth and repeat channels
	// This initial run is used to determine the het parameters for the subsequent iterations
	//
	ModelParameters modelParams = new ModelParameters(model);

	// Calculate the initial depth parameters
	Statistics depthStats = calculateInitialDepth(symbols, 0.10);

	// Set initial depth params
	modelParams.setDepthParameters(depthStats.mean, depthStats.sd, ZERO_DEPTH_PRIOR, depthStats.mean, depthStats.sd);

	// Average hit priors
	modelParams.setRepeatParameters(UNIQUE_RMEAN_PRIOR, UNIQUE_RSD_PRIOR, ZERO_RMEAN_PRIOR, ZERO_RSD_PRIOR, REPEAT_RMEAN_PRIOR, REPEAT_RSD_PRIOR);
	return modelParams;
}

//
// Create model parameters from a previous classification
//
ModelParameters generateParameters(MarkovModel model, SymbolArray[State] stateSymbols)
{
	ModelParameters parameters = new ModelParameters(model);
	State[] states = model.getStateList();
	StateSummary summary = new StateSummary(model, stateSymbols);

	// Strong prior on the zero state parameters as sequence gaps can pull the values down
	real zeroPrior = 5000;

	//
	// Estimate depth by calculating the mean/stddev over the normal state and using it
	// to set the depth parameters for all the states
	//
	Statistics zeroStats = summary.getStats(SNAME_ZERO, DEPTH_COL_NAME);
	real zeroDepthMean = calculateWeightedValue(zeroStats.numDataPoints, zeroPrior, zeroStats.mean, ZERO_DEPTH_PRIOR);

	Statistics depthStats = summary.getStats(SNAME_NORMAL, DEPTH_COL_NAME);
	Statistics repeatDepth = summary.getStats(SNAME_RPTCMN, DEPTH_COL_NAME);
	parameters.setDepthParameters(depthStats.mean, depthStats.sd, zeroDepthMean, depthStats.mean, depthStats.sd);
	
	//
	// Set the het parameters
	// N/Z/R/C0.5 all get the background het, estimated from the normal classification
	//
	real hetPriorWeight = 10;
	real defaultHetMean = 2.0f;	
	real backgroundHet = summary.getStats(SNAME_NORMAL, HET_COL_NAME).mean;

	// Calculate C2/C3 parameters
	Statistics c2HetStats = summary.getStats(SNAME_C2, HET_COL_NAME);
	Statistics c3HetStats = summary.getStats(SNAME_C3, HET_COL_NAME);

	real c2HetMean = calculateWeightedValue(c2HetStats.numDataPoints, hetPriorWeight, c2HetStats.mean, defaultHetMean);
	real c2HetSD = calculateWeightedValue(c2HetStats.numDataPoints, hetPriorWeight, c2HetStats.sd, defaultHetMean);

	real c3HetMean = calculateWeightedValue(c3HetStats.numDataPoints, hetPriorWeight, c3HetStats.mean, defaultHetMean);
	real c3HetSD = calculateWeightedValue(c3HetStats.numDataPoints, hetPriorWeight, c3HetStats.sd, defaultHetMean);
	parameters.setHetParameters(backgroundHet, c2HetMean, c2HetSD, c3HetMean, c3HetSD);

	//
	// Set the repeat parameters to the mean value for the normal/repeat/zero states
	// The C0.5/N/C2/C3 all use the estimate from the N state
	Statistics repeatStats = summary.getStats(SNAME_RPTCMN, AVGHIT_COL_NAME);
	Statistics zeroRepeatStats = summary.getStats(SNAME_ZERO, AVGHIT_COL_NAME);
	Statistics uniqueRepeatStats = summary.getStats(SNAME_NORMAL, AVGHIT_COL_NAME);

	parameters.setRepeatParameters(uniqueRepeatStats.mean, uniqueRepeatStats.sd, zeroRepeatStats.mean, zeroRepeatStats.sd, repeatStats.mean, repeatStats.sd);
	return parameters;
}

//
// Calculate the weighted value of a parameter based on a prior estimate and evidence
//
real calculateWeightedValue(real evidenceWeight, real priorWeight, real evidenceValue, real priorValue)
{
	real sum = evidenceWeight + priorWeight;
	real v = (evidenceWeight / sum) * evidenceValue + (priorWeight / sum) * priorValue;
	return v;
}

//
// Calculate the length of stay in each state based on the viterbi output
//
void updateTransitions(MarkovModel model, State[] viterbiPath, real smoothingFactor)
{
	RealArray[State] stateLengths;
	State[] stateList = model.getStateList();

	foreach(s; stateList)
	{
		stateLengths[s] = new RealArray;
	}

	State prev = model.getTerminalState();
	real currLen = 1.0;
	foreach(s; viterbiPath)
	{
		if(s == prev)
		{
			currLen++;
		}
		else
		{
			//Stdout.formatln("End of run {} {}", prev, currLen);
			stateLengths[prev].append(currLen);
			currLen = 1.0; 
		}
		prev = s;
	}

	// Add the last
	stateLengths[prev].append(currLen);

	Stdout("\nState length summary:\n");
	// Calculate the mean time in each state
	foreach(s; stateList)
	{
		if(s == model.getTerminalState())
			continue;

		Statistics stats = computeStats(stateLengths[s].toArray());
		Stdout.formatln("State {} n: {} mean: {} sd: {} sum: {}", s.toString(), stats.numDataPoints, stats.mean, stats.sd, stats.mean*stats.numDataPoints);
		
		real mean;
		if(stats.numDataPoints == 0)
		{
			mean = 10;
		}
		else
		{
			mean = stats.mean * smoothingFactor;
		}

		DiscreteDistribution transitions = model.getTransitionWeights(s);
		real p_self = 1 - (1 / mean);
		real p_anyOther = 1 - p_self;

		int numOthers = 0;
		foreach(l; stateList)
		{
			if(model.hasTransition(s, l) && s != l)
			{
				numOthers++;
			}
		}

		real p_someOther = p_anyOther / numOthers;

		foreach(l; stateList)
		{
			if(model.hasTransition(s,l))
			{
				if(s == l)
					transitions.setWeight(l, p_self);
				else
					transitions.setWeight(l, p_someOther);
			}
		}
	}
}


// Calculate the initial estimate for the mean depth of the 1-copy state
// For the initial mean depth estimate the trimmed mean is used
Statistics calculateInitialDepth(Symbol[] symbols, real trim_percent)
{
	// Extract the depth column
	int length = symbols.length; 

	real[] depthData;
	depthData.length = length;
	
	int idx = 0;	
	foreach(s; symbols)
	{
		JointSymbol js = cast(JointSymbol)s;
		depthData[idx++] = js.depthSymbol.value;		
	}
	
	depthData.sort;

	// Calculate how much to trim
	int trimmed_length = rndint((cast(real)length * (1.0f- trim_percent)));
	int trim_start = rndint(cast(real)length * (trim_percent/2));	
	
	// Perform the trim
	real[] trimmed;
	trimmed.length = trimmed_length;
	int count = 0;
	for(int i = 0; i < depthData.length; ++i)
	{
		if(i > trim_start && count < trimmed_length)
		{
			trimmed[count++] = depthData[i];
		}
	}

	// Compute the mean over the trimmed data as the estimate
	Statistics stats = computeStats(trimmed);
	
	// Set the initial model parameters
	return stats;
}

