//-----------------------------------------------
// Copyright 2009 Wellcome Trust Sanger Institute
// Written by Jared Simpson (js18@sanger.ac.uk)
// Released under the GPL license
//-----------------------------------------------
import Alphabet;
import Distribution;
import State;
import MarkovModel;
import HmmAlgorithms;
import CNVModel;
import tango.io.Stdout;
import Integer = tango.text.convert.Integer;
import Float = tango.text.convert.Float;
import tango.io.stream.FileStream;
import tango.text.Util;
import tango.io.stream.LineStream;
import tango.util.ArgParser;
import tango.math.Math;
import tango.util.collection.ArraySeq;

class JointSymbol : Symbol
{
	this(RealSymbol hs, RealSymbol ds, RealSymbol hitsym)
	{
		hetSymbol = hs;
		depthSymbol = ds;
		hitSymbol = hitsym;
	}

	char[] toString() { return hetSymbol.toString() ~ "," ~ depthSymbol.toString() ~ "," ~ hitSymbol.toString(); }


	RealSymbol hetSymbol;
	RealSymbol depthSymbol;
	RealSymbol hitSymbol;
}

//
// The JointCNVDistribution is the triplet of values (het rate, depth, number of hits) emitted by the model
// The three values are assumed to be independent
//
class JointCNVDistribution : IDistribution
{
	this(char[] name, IDistribution hetD, IDistribution depthD, IDistribution hitD)
	{
		m_hetDist = hetD;
		m_depthDist = depthD;
		m_hitDist = hitD;
		m_name = name;
	}

	// Get the weight of the given symbol (from the PDF or PMF)
	real getWeight(Symbol s)
	{
		if(s.classinfo != JointSymbol.classinfo)
		{
			throw new Exception("Invalid symbol type to JointCNVDistribution: " ~ s.classinfo.name);
		}
		
		JointSymbol js = cast(JointSymbol)s;

		real p = 1.0;
		p *= m_useDepth ? m_depthDist.getWeight(js.depthSymbol) : 1.0;
		p *= m_useHet ? m_hetDist.getWeight(js.hetSymbol) : 1.0;
		p *= m_useRepeat ? m_hitDist.getWeight(js.hitSymbol) : 1.0;
		//printWeight(s);
		return p;
	}

	// Get the probability (from the CDF) that X is less than x
	real getProbLessThan(Symbol s)
	{
		if(s.classinfo != JointSymbol.classinfo)
		{
			throw new Exception("Invalid symbol type to JointCNVDistribution: " ~ s.classinfo.name);
		}

		JointSymbol js = cast(JointSymbol)s;

		real p = 1.0;
		p *= m_useDepth ? m_depthDist.getProbLessThan(js.depthSymbol) : 1.0;
		p *= m_useHet ? m_hetDist.getProbLessThan(js.hetSymbol) : 1.0;
		p *= m_useRepeat ? m_hitDist.getProbLessThan(js.hitSymbol) : 1.0;
		return p;		
	}

	void printWeight(Symbol s)
	{
		if(s.classinfo != JointSymbol.classinfo)
		{
			throw new Exception("Invalid symbol type to JointCNVDistribution: " ~ s.classinfo.name);
		}
		
		JointSymbol js = cast(JointSymbol)s;

		real d_w = m_useDepth ? m_depthDist.getWeight(js.depthSymbol) : 1.0;
		real h_w = m_useHet ? m_hetDist.getWeight(js.hetSymbol) : 1.0;
		real r_w = m_useRepeat ? m_hitDist.getWeight(js.hitSymbol) : 1.0;
		real p = d_w * h_w * r_w;
		Stdout.format("Emission Probs({0}) [{1} -> {2:F6}] [{3} -> {4:F6}] [{5} -> {6:F6} = {7:F6}]\n",
					  m_name, js.depthSymbol.toString(), d_w, js.hetSymbol.toString(), h_w, js.hitSymbol.toString(), r_w, p);
	}

	// Sample a symbol from the distribution
	Symbol sample()
	{
		throw new Exception("Sampling not implemented for joint cnv");
		return new Symbol("");		
	}

	// Does the distribution contain this symbol
	bool contains(Symbol s)
	{
		return true;
	}

	char[] toString()
	{
		char[] outstr;
		outstr = "DepthDist: " ~ m_depthDist.toString() ~ " HetDist: " ~ m_hetDist.toString() ~ " HitDist: " ~ m_hitDist.toString();
		return outstr;
	}

	static void enableDepth() { m_useDepth = true; }
	static void enableHet() { m_useHet = true; }
	static void enableRepeat() { m_useRepeat = true; }

	static void disableDepth() { m_useDepth = false; }
	static void disableHet() { m_useHet = false; }
	static void disableRepeat() { m_useRepeat = false; }

	IDistribution m_hetDist;
	IDistribution m_depthDist;
	IDistribution m_hitDist;
	char[] m_name;

	static bool m_useDepth = true;
	static bool m_useHet = true;
	static bool m_useRepeat = true;
}	

